import enum
from typing import Callable, ClassVar, Dict, Hashable, Iterable, Set, Union

import _abc
import graph_scheduler.time
import pint
from _typeshed import Incomplete

__all__ = ['Operation', 'ConditionError', 'ConditionSet', 'ConditionBase', 'Condition', 'AbsoluteCondition', 'While', 'When', 'WhileNot', 'Always', 'Never', 'CompositeCondition', 'All', 'Any', 'And', 'Or', 'Not', 'NWhen', 'TimeInterval', 'TimeTermination', 'BeforeTimeStep', 'AtTimeStep', 'AfterTimeStep', 'AfterNTimeSteps', 'BeforePass', 'AtPass', 'AfterPass', 'AfterNPasses', 'EveryNPasses', 'BeforeTrial', 'AtTrial', 'AfterTrial', 'AfterNTrials', 'AtRun', 'AfterRun', 'AfterNRuns', 'BeforeNCalls', 'AtNCalls', 'AfterCall', 'AfterNCalls', 'AfterNCallsCombined', 'EveryNCalls', 'JustRan', 'AllHaveRun', 'WhenFinished', 'WhenFinishedAny', 'WhenFinishedAll', 'AtTrialStart', 'AtTrialNStart', 'AtRunStart', 'AtRunNStart', 'Threshold', 'GraphStructureCondition', 'CustomGraphStructureCondition', 'BeforeNodes', 'BeforeNode', 'WithNode', 'AfterNodes', 'AfterNode', 'AddEdgeTo', 'RemoveEdgeFrom']


SubjectOperation = Union['Operation', str, Dict[Hashable, Union['Operation', str]]]
ConditionSetDict = Dict[Hashable, Union['ConditionBase', Iterable['ConditionBase']]]
GraphDependencyDict = Dict[Hashable, Set[Hashable]]


class Operation(enum.Enum):

    """
    Used in conjunction with `GraphStructureCondition` to indicate how a
    set of source nodes (**S** below) should be combined with a set of
    comparison nodes (**C** below) to produce a result set. Many
    Operations correspond to standard set operations.

    Each enum item can be called with a source set and comparison set as
    arguments to produce the result set.

    Attributes:
        KEEP: Returns **S**

        REPLACE: Returns **C**

        DISCARD: Returns the empty set

        INTERSECTION: Returns the set of items that are in both **S**
            and **C**

        UNION: Returns the set of items in either **S** or **C**

        MERGE: Returns the set of items in either **S** or **C**

        DIFFERENCE: Returns the set of items in **S** but not **C**

        INVERSE_DIFFERENCE: Returns the set of items in **C** but not
            **S**

        SYMMETRIC_DIFFERENCE: Returns the set of items that are in one
            of **S** or **C** but not both

    """
    _member_names_: ClassVar[list] = ...
    _member_map_: ClassVar[dict] = ...
    _member_type_: ClassVar[type[object]] = ...
    _value2member_map_: ClassVar[dict] = ...
    KEEP: ClassVar[Operation] = ...
    REPLACE: ClassVar[Operation] = ...
    DISCARD: ClassVar[Operation] = ...
    INTERSECTION: ClassVar[Operation] = ...
    UNION: ClassVar[Operation] = ...
    MERGE: ClassVar[Operation] = ...
    DIFFERENCE: ClassVar[Operation] = ...
    INVERSE_DIFFERENCE: ClassVar[Operation] = ...
    SYMMETRIC_DIFFERENCE: ClassVar[Operation] = ...
    def __call__(self, source_neighbors: set[Hashable], comparison_neighbors: set[Hashable]) -> set[Hashable]:
        """
        Returns the set resulting from applying an `Operation` on
        **source_neighbors** and **comparison_neighbors**

        Args:
            source_neighbors (Set[Hashable])
            comparison_neighbors (Set[Hashable])

        Returns:
            Set[Hashable]
        """
    @classmethod
    def __init__(cls, value) -> None: ...

class ConditionError(Exception):
    __init__: ClassVar[wrapper_descriptor] = ...

class ConditionSet:

    """Used in conjunction with a `Scheduler <graph_scheduler.scheduler.Scheduler>` to store the `Conditions <Condition>` associated with a node.

    Arguments
    ---------

    *condition_sets
        each item is a dict or ConditionSet mapping nodes to one or more
        conditions to be added via `ConditionSet.add_condition_set`

    conditions
        a dict or ConditionSet mapping nodes to one or more conditions
        to be added via `ConditionSet.add_condition_set`. Maintained for
        backwards compatibility with versions 1.x

    Attributes
    ----------

    conditions : Dict[Hashable: Union[ConditionBase, Iterable[ConditionBase]]]
        the key of each entry is a node, and its value is a condition
        associated
        with that node.  Conditions can be added to the
        ConditionSet using the ConditionSet's `add_condition` method.

    conditions_basic : Dict[Hashable: `Condition <graph_scheduler.condition.Condition>`]
        a dict mapping nodes to their single `basic Conditions
        <graph_scheduler.condition.Condition>`

    conditions_structural : Dict[Hashable: List[`GraphStructureCondition`]]
        a dict mapping nodes to their `graph structure Conditions
        <GraphStructureCondition>`

    structural_condition_order : List[`GraphStructureCondition`]
        a list storing all `GraphStructureCondition` s in this
        ConditionSet in the order in which they were added (and will be
        applied to a `Scheduler`)

    """
    def __init__(self, *condition_sets: ConditionSetDict, conditions: ConditionSetDict = ...) -> None: ...
    def __contains__(self, item) -> bool: ...
    def __iter__(self): ...
    def __getitem__(self, key): ...
    def __setitem__(self, key, value) -> None: ...
    def add_condition(self, owner: Hashable, condition: ConditionBase):
        """
        Adds a `basic <graph_scheduler.condition.Condition>` or `graph
        structure <GraphStructureCondition>` Condition to the
        ConditionSet.

        If **condition** is basic, it will overwrite the current basic
        Condition for **owner**, if present. If you want to add multiple
        basic Conditions to a single owner, instead add a single
        `Composite Condition <Conditions_Composite>` to accurately
        specify the desired behavior.

        Arguments
        ---------

        owner : node
            specifies the node with which the **condition** should be associated. **condition**
            will govern the execution behavior of **owner**

        condition : ConditionBase
            specifies the condition associated with **owner** to be
            added to the ConditionSet.
        """
    def remove_condition(self, owner_or_condition: Hashable | ConditionBase) -> ConditionBase | None:
        """
        Removes the condition specified as or owned by
        **owner_or_condition**.

        Args:
            owner_or_condition (Union[Hashable, `ConditionBase`]):
                Either a condition or the owner of a condition

        Returns:
            The condition removed, or None if no condition removed

        Raises:
            ConditionError:
                - when **owner_or_condition** is an owner and it owns
                  multiple conditions
                - when **owner_or_condition** is a condition and its
                  owner is None
        """
    def add_condition_set(self, conditions: ConditionSet | ConditionSetDict):
        """
        Adds a set of `basic <graph_scheduler.condition.Condition>` or
        `graph structure <GraphStructureCondition>` Conditions (in the
        form of a dict or another ConditionSet) to the ConditionSet.

        Any basic Condition added here will overwrite the current basic
        Condition for a given owner, if present. If you want to add
        multiple basic Conditions to a single owner, instead add a
        single `Composite Condition <Conditions_Composite>` to
        accurately specify the desired behavior.

        Arguments
        ---------

        conditions
            specifies collection of Conditions to be added to this ConditionSet,

            if a dict is provided:
                each entry should map an owner node (the node whose
                execution behavior will be governed) to a `Condition
                <graph_scheduler.condition.Condition>` or
                `GraphStructureCondition`, or an iterable of them.

        """
    @property
    def conditions(self): ...

class ConditionBase:

    """
    Abstract base class for `basic conditions
    <graph_scheduler.condition.Condition>` and `graph structure
    conditions <GraphStructureCondition>`

    Attributes:
        owner (Hashable):
            the node with which the Condition is associated, and the
            execution of which it determines.

    """
    owner: Incomplete
    def __init__(self, _owner: Hashable = ..., **kwargs) -> None: ...

class Condition(ConditionBase):

    """
    Used in conjunction with a :class:`Scheduler` to specify the condition under which a node should be
    allowed to execute.

    Arguments
    ---------

    func : callable
        specifies function to be called when the Condition is evaluated, to determine whether it is currently satisfied.

    args : *args
        specifies formal arguments to pass to `func` when the Condition is evaluated.

    kwargs : **kwargs
        specifies keyword arguments to pass to `func` when the Condition is evaluated.
    """
    def __init__(self, func, *args, **kwargs) -> None: ...
    def is_satisfied(self, *args, execution_id: Incomplete | None = ..., **kwargs):
        """
        the function called to determine satisfaction of this Condition.

        Arguments
        ---------
        args : *args
            specifies additional formal arguments to pass to `func` when the Condition is evaluated.
            these are appended to the **args** specified at instantiation of this Condition

        kwargs : **kwargs
            specifies additional keyword arguments to pass to `func` when the Condition is evaluated.
            these are added to the **kwargs** specified at instantiation of this Condition

        Returns
        -------
            True - if the Condition is satisfied
            False - if the Condition is not satisfied
        """
    @property
    def absolute_intervals(self): ...
    @property
    def absolute_fixed_points(self): ...
    @property
    def is_absolute(self): ...

class AbsoluteCondition(Condition):
    def __init__(self, func, *args, **kwargs) -> None: ...
    @property
    def is_absolute(self): ...

class _DependencyValidation: ...
class While(ConditionBase):

    """
    Used in conjunction with a :class:`Scheduler` to specify the condition under which a node should be
    allowed to execute.

    Arguments
    ---------

    func : callable
        specifies function to be called when the Condition is evaluated, to determine whether it is currently satisfied.

    args : *args
        specifies formal arguments to pass to `func` when the Condition is evaluated.

    kwargs : **kwargs
        specifies keyword arguments to pass to `func` when the Condition is evaluated.
    """
    def __init__(self, func, *args, **kwargs) -> None: ...
    def is_satisfied(self, *args, execution_id: Incomplete | None = ..., **kwargs):
        """
        the function called to determine satisfaction of this Condition.

        Arguments
        ---------
        args : *args
            specifies additional formal arguments to pass to `func` when the Condition is evaluated.
            these are appended to the **args** specified at instantiation of this Condition

        kwargs : **kwargs
            specifies additional keyword arguments to pass to `func` when the Condition is evaluated.
            these are added to the **kwargs** specified at instantiation of this Condition

        Returns
        -------
            True - if the Condition is satisfied
            False - if the Condition is not satisfied
        """
    @property
    def absolute_intervals(self): ...
    @property
    def absolute_fixed_points(self): ...
    @property
    def is_absolute(self): ...

class When(ConditionBase):

    """
    Used in conjunction with a :class:`Scheduler` to specify the condition under which a node should be
    allowed to execute.

    Arguments
    ---------

    func : callable
        specifies function to be called when the Condition is evaluated, to determine whether it is currently satisfied.

    args : *args
        specifies formal arguments to pass to `func` when the Condition is evaluated.

    kwargs : **kwargs
        specifies keyword arguments to pass to `func` when the Condition is evaluated.
    """
    def __init__(self, func, *args, **kwargs) -> None: ...
    def is_satisfied(self, *args, execution_id: Incomplete | None = ..., **kwargs):
        """
        the function called to determine satisfaction of this Condition.

        Arguments
        ---------
        args : *args
            specifies additional formal arguments to pass to `func` when the Condition is evaluated.
            these are appended to the **args** specified at instantiation of this Condition

        kwargs : **kwargs
            specifies additional keyword arguments to pass to `func` when the Condition is evaluated.
            these are added to the **kwargs** specified at instantiation of this Condition

        Returns
        -------
            True - if the Condition is satisfied
            False - if the Condition is not satisfied
        """
    @property
    def absolute_intervals(self): ...
    @property
    def absolute_fixed_points(self): ...
    @property
    def is_absolute(self): ...

class WhileNot(Condition):

    """
    WhileNot

    Parameters:

        func : callable
            specifies function to be called when the Condition is evaluated, to determine whether it is currently satisfied.

        args : *args
            specifies formal arguments to pass to `func` when the Condition is evaluated.

        kwargs : **kwargs
            specifies keyword arguments to pass to `func` when the Condition is evaluated.

    Satisfied when:

        - **func** is False

    """
    def __init__(self, func, *args, **kwargs) -> None: ...

class Always(Condition):

    """Always

    Parameters:

        none

    Satisfied when:

        - always satisfied.

    """
    def __init__(self) -> None: ...

class Never(Condition):

    """Never

    Parameters:

        none

    Satisfied when:

        - never satisfied.
    """
    def __init__(self) -> None: ...

class CompositeCondition(Condition):
    owner: Incomplete
    def __init__(self, func, *args, **kwargs) -> None: ...
    @property
    def absolute_intervals(self): ...
    @property
    def absolute_fixed_points(self): ...
    @property
    def is_absolute(self): ...

class All(CompositeCondition):

    """All

    Parameters:

        args: one or more `Conditions <Condition>`

    Satisfied when:

        - all of the Conditions in args are satisfied.

    Notes:

        - To initialize with a list (for example)::

            conditions = [AfterNCalls(node, 5) for node in node_list]

          unpack the list to supply its members as args::

           composite_condition = All(*conditions)

    """
    def __init__(self, *args, **dependencies) -> None: ...
    def satis(self, *conds, **kwargs): ...

class Any(CompositeCondition):

    """Any

    Parameters:

        args: one or more `Conditions <Condition>`

    Satisfied when:

        - one or more of the Conditions in **args** is satisfied.

    Notes:

        - To initialize with a list (for example)::

            conditions = [AfterNCalls(node, 5) for node in node_list]

          unpack the list to supply its members as args::

           composite_condition = Any(*conditions)

    """
    def __init__(self, *args, **dependencies) -> None: ...
    def satis(self, *conds, **kwargs): ...

class And(CompositeCondition):

    """All

    Parameters:

        args: one or more `Conditions <Condition>`

    Satisfied when:

        - all of the Conditions in args are satisfied.

    Notes:

        - To initialize with a list (for example)::

            conditions = [AfterNCalls(node, 5) for node in node_list]

          unpack the list to supply its members as args::

           composite_condition = All(*conditions)

    """
    def __init__(self, *args, **dependencies) -> None: ...
    def satis(self, *conds, **kwargs): ...

class Or(CompositeCondition):

    """Any

    Parameters:

        args: one or more `Conditions <Condition>`

    Satisfied when:

        - one or more of the Conditions in **args** is satisfied.

    Notes:

        - To initialize with a list (for example)::

            conditions = [AfterNCalls(node, 5) for node in node_list]

          unpack the list to supply its members as args::

           composite_condition = Any(*conditions)

    """
    def __init__(self, *args, **dependencies) -> None: ...
    def satis(self, *conds, **kwargs): ...

class Not(Condition):

    """Not

    Parameters:

        condition(Condition): a `Condition`

    Satisfied when:

        - **condition** is not satisfied.

    """
    owner: Incomplete
    def __init__(self, condition) -> None: ...

class NWhen(Condition):

    """NWhen

    Parameters:

        condition(Condition): a `Condition`

        n(int): the maximum number of times this condition will be satisfied

    Satisfied when:

        - the first **n** times **condition** is satisfied upon evaluation

    """
    owner: Incomplete
    def __init__(self, condition, n: int = ...) -> None: ...
    def satis(self, condition, n, *args, scheduler: Incomplete | None = ..., execution_id: Incomplete | None = ..., **kwargs): ...

class TimeInterval(AbsoluteCondition):

    """TimeInterval

    Attributes:

        repeat
            the interval between *unit*s where this condition can be
            satisfied

        start
            the time at/after which this condition can be
            satisfied

        end
            the time at/fter which this condition can be
            satisfied

        unit
            the `pint.Unit` to use for scalar values of *repeat*,
            *start*, and *end*

        start_inclusive
            if True, *start* allows satisfaction exactly at the time
            corresponding to *start*. if False, satisfaction can occur
            only after *start*

        end_inclusive
            if True, *end* allows satisfaction exactly until the time
            corresponding to *end*. if False, satisfaction can occur
            only before *end*


    Satisfied when:

        Every *repeat* units of time at/after *start* and before/through
        *end*

    Notes:

        Using a `TimeInterval` as a
        `termination Condition <Scheduler_Termination_Conditions>` may
        result in unexpected behavior. The user may be inclined to
        create **TimeInterval(end=x)** to terminate at time **x**, but
        this will do the opposite and be true only and always until time
        **x**, terminating at any time before **x**. If in doubt, use
        `TimeTermination` instead.

        If the scheduler is not set to `exact_time_mode = True`,
        *start_inclusive* and *end_inclusive* may not behave as
        expected. See `Scheduler_Exact_Time` for more info.
    """
    def __init__(self, repeat: int | str | pint.Quantity = ..., start: int | str | pint.Quantity = ..., end: int | str | pint.Quantity = ..., unit: str | pint.Unit = ..., start_inclusive: bool = ..., end_inclusive: bool = ...) -> None: ...
    @property
    def absolute_intervals(self): ...
    @property
    def absolute_fixed_points(self): ...

class TimeTermination(AbsoluteCondition):

    """TimeTermination

    Attributes:

        t
            the time at/after which this condition is satisfied

        unit
            the `pint.Unit` to use for scalar values of *t*, *start*,
            and *end*

        start_inclusive
            if True, the condition is satisfied exactly at the time
            corresponding to *t*. if False, satisfaction can occur
            only after *t*

    Satisfied when:

        At/After time *t*
    """
    def __init__(self, t: int | str | pint.Quantity, inclusive: bool = ..., unit: str | pint.Unit = ...) -> None: ...
    @property
    def absolute_fixed_points(self): ...

class BeforeTimeStep(Condition):

    """BeforeTimeStep

    Parameters:

        n(int): the 'TIME_STEP' before which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `TIME_STEP`\\ s (default: TimeScale.TRIAL)

    Satisfied when:

        - at most n-1 `TIME_STEP`\\ s have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `TIME_STEP` is 0, the second `TIME_STEP` is 1, etc.);
          so, `BeforeTimeStep(2)` is satisfied at `TIME_STEP` 0 and `TIME_STEP` 1.

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AtTimeStep(Condition):

    """AtTimeStep

    Parameters:

        n(int): the `TIME_STEP` at which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `TIME_STEP`\\ s (default: TimeScale.TRIAL)

    Satisfied when:

        - exactly n `TIME_STEP`\\ s have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first 'TIME_STEP' is pass 0, the second 'TIME_STEP' is 1, etc.);
          so, `AtTimeStep(1)` is satisfied when a single `TIME_STEP` (`TIME_STEP` 0) has occurred, and `AtTimeStep(2)` is satisfied
          when two `TIME_STEP`\\ s have occurred (`TIME_STEP` 0 and `TIME_STEP` 1), etc..

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AfterTimeStep(Condition):

    """AfterTimeStep

    Parameters:

        n(int): the `TIME_STEP` after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `TIME_STEP`\\ s (default: TimeScale.TRIAL)

    Satisfied when:

        - at least n+1 `TIME_STEP`\\ s have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScals are zero-indexed (that is, the first `TIME_STEP` is 0, the second `TIME_STEP` is 1, etc.); so,
          `AfterTimeStep(1)` is satisfied after `TIME_STEP` 1 has occurred and thereafter (i.e., in `TIME_STEP`\\ s 2, 3, 4, etc.).

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AfterNTimeSteps(Condition):

    """AfterNTimeSteps

    Parameters:

        n(int): the number of `TIME_STEP`\\ s after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `TIME_STEP`\\ s (default: TimeScale.TRIAL)


    Satisfied when:

        - at least n `TIME_STEP`\\ s have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class BeforePass(Condition):

    """BeforePass

    Parameters:

        n(int): the 'PASS' before which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `PASS`\\ es (default: TimeScale.TRIAL)

    Satisfied when:

        - at most n-1 `PASS`\\ es have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `PASS` is 0, the second `PASS` is 1, etc.);
          so, `BeforePass(2)` is satisfied at `PASS` 0 and `PASS` 1.

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AtPass(Condition):

    """AtPass

    Parameters:

        n(int): the `PASS` at which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `PASS`\\ es (default: TimeScale.TRIAL)

    Satisfied when:

        - exactly n `PASS`\\ es have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first 'PASS' is pass 0, the second 'PASS' is 1, etc.);
          so, `AtPass(1)` is satisfied when a single `PASS` (`PASS` 0) has occurred, and `AtPass(2)` is satisfied
          when two `PASS`\\ es have occurred (`PASS` 0 and `PASS` 1), etc..

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AfterPass(Condition):

    """AfterPass

    Parameters:

        n(int): the `PASS` after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `PASS`\\ es (default: TimeScale.TRIAL)

    Satisfied when:

        - at least n+1 `PASS`\\ es have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `PASS` is 0, the second `PASS` is 1, etc.); so,
          `AfterPass(1)` is satisfied after `PASS` 1 has occurred and thereafter (i.e., in `PASS`\\ es 2, 3, 4, etc.).

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AfterNPasses(Condition):

    """AfterNPasses

    Parameters:

        n(int): the number of `PASS`\\ es after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `PASS`\\ es (default: TimeScale.TRIAL)


    Satisfied when:

        - at least n `PASS`\\ es have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class EveryNPasses(Condition):

    """EveryNPasses

    Parameters:

        n(int): the frequency of passes with which this condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `PASS`\\ es (default: TimeScale.TRIAL)

    Satisfied when:

        - `PASS` 0

        - the specified number of `PASS`\\ es that has occurred within a unit of time (at the `TimeScale` specified by
          **time_scale**) is evenly divisible by n.

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class BeforeTrial(Condition):

    """BeforeTrial

    Parameters:

        n(int): the `TRIAL <TimeScale.TRIAL>` before which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `TRIAL <TimeScale.TRIAL>`\\ s
        (default: TimeScale.RUN)

    Satisfied when:

        - at most n-1 `TRIAL <TimeScale.TRIAL>`\\ s have occurred within one unit of time at the `TimeScale`
          specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `TRIAL <TimeScale.TRIAL>` is 0, the second
          `TRIAL <TimeScale.TRIAL>` is 1, etc.); so, `BeforeTrial(2)` is satisfied at `TRIAL <TimeScale.TRIAL>` 0
          and `TRIAL <TimeScale.TRIAL>` 1.

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AtTrial(Condition):

    """AtTrial

    Parameters:

        n(int): the `TRIAL <TimeScale.TRIAL>` at which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `TRIAL <TimeScale.TRIAL>`\\ s
        (default: TimeScale.RUN)

    Satisfied when:

        - exactly n `TRIAL <TimeScale.TRIAL>`\\ s have occurred within one unit of time at the `TimeScale`
          specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `TRIAL <TimeScale.TRIAL>` is 0,
          the second `TRIAL <TimeScale.TRIAL>` is 1, etc.); so, `AtTrial(1)` is satisfied when one
          `TRIAL <TimeScale.TRIAL>` (`TRIAL <TimeScale.TRIAL>` 0) has already occurred.

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AfterTrial(Condition):

    """AfterTrial

    Parameters:

        n(int): the `TRIAL <TimeScale.TRIAL>` after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `TRIAL <TimeScale.TRIAL>`\\ s.
        (default: TimeScale.RUN)

    Satisfied when:

        - at least n+1 `TRIAL <TimeScale.TRIAL>`\\ s have occurred within one unit of time at the `TimeScale`
          specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `TRIAL <TimeScale.TRIAL>` is 0, the second
        `TRIAL <TimeScale.TRIAL>` is 1, etc.); so,  `AfterPass(1)` is satisfied after `TRIAL <TimeScale.TRIAL>` 1
        has occurred and thereafter (i.e., in `TRIAL <TimeScale.TRIAL>`\\ s 2, 3, 4, etc.).

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AfterNTrials(Condition):

    """AfterNTrials

    Parameters:

        n(int): the number of `TRIAL <TimeScale.TRIAL>`\\ s after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `TRIAL <TimeScale.TRIAL>`\\ s
        (default: TimeScale.RUN)

    Satisfied when:

        - at least n `TRIAL <TimeScale.TRIAL>`\\ s have occured within one unit of time at the `TimeScale`
          specified by **time_scale**.

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AtRun(Condition):

    """AtRun

    Parameters:

        n(int): the `RUN` at which the Condition is satisfied

    Satisfied when:

        - exactly n `RUN`\\ s have occurred.

    Notes:
        - `RUN`\\ s are managed by the environment         using the Scheduler (e.g.         `end_environment_sequence <Scheduler.end_environment_sequence>`        ) and are not automatically updated by this package.

    """
    def __init__(self, n) -> None: ...

class AfterRun(Condition):

    """AfterRun

    Parameters:

        n(int): the `RUN` after which the Condition is satisfied

    Satisfied when:

        - at least n+1 `RUN`\\ s have occurred.

    Notes:
        - `RUN`\\ s are managed by the environment         using the Scheduler (e.g.         `end_environment_sequence <Scheduler.end_environment_sequence>`        ) and are not automatically updated by this package.

    """
    def __init__(self, n) -> None: ...

class AfterNRuns(Condition):

    """AfterNRuns

    Parameters:

        n(int): the number of `RUN`\\ s after which the Condition is satisfied

    Satisfied when:

        - at least n `RUN`\\ s have occured.

    Notes:
        - `RUN`\\ s are managed by the environment         using the Scheduler (e.g.         `end_environment_sequence <Scheduler.end_environment_sequence>`        ) and are not automatically updated by this package.

    """
    def __init__(self, n) -> None: ...

class BeforeNCalls(_DependencyValidation, Condition):

    """BeforeNCalls

    Parameters:

        dependency (Hashable):  the node on which the Condition depends

        n(int): the number of executions of **dependency** before which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.TRIAL)

    Satisfied when:

        - the node specified in **dependency** has executed at most n-1 times
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, dependency, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AtNCalls(_DependencyValidation, Condition):

    """AtNCalls

    Parameters:

        dependency (Hashable):  the node on which the Condition depends

        n(int): the number of executions of **dependency** at which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.TRIAL)

    Satisfied when:

        - the node specified in **dependency** has executed exactly n times
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, dependency, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AfterCall(_DependencyValidation, Condition):

    """AfterCall

    Parameters:

        dependency (Hashable):  the node on which the Condition depends

        n(int): the number of executions of **dependency** after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.TRIAL)

    Satisfied when:

        - the node specified in **dependency** has executed at least n+1 times
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, dependency, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AfterNCalls(_DependencyValidation, Condition):

    """AfterNCalls

    Parameters:

        dependency (Hashable):  the node on which the Condition depends

        n(int): the number of executions of **dependency** after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.TRIAL)

    Satisfied when:

        - the node specified in **dependency** has executed at least n times
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, dependency, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AfterNCallsCombined(_DependencyValidation, Condition):

    """AfterNCallsCombined

    Parameters:

        *dependencies (Hashable):  one or more nodes on which the Condition depends

        n(int): the number of combined executions of all nodes specified in **dependencies** after which the
        Condition is satisfied (default: None)

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.TRIAL)


    Satisfied when:

        - there have been at least n+1 executions among all of the nodes specified in **dependencies**
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, *dependencies, n: Incomplete | None = ..., time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class EveryNCalls(_DependencyValidation, Condition):

    '''EveryNCalls

    Parameters:

        dependency (Hashable):  the node on which the Condition depends

        n(int): the frequency of executions of **dependency** at which the Condition is satisfied


    Satisfied when:

        - the node specified in **dependency** has executed at least n times since the last time the
          Condition\'s owner executed.


    Notes:

        - scheduler\'s count of each other node that is "useable" by the node is reset to 0 when the
          node runs

    '''
    def __init__(self, dependency, n) -> None: ...

class JustRan(_DependencyValidation, Condition):

    """JustRan

    Parameters:

        dependency (Hashable):  the node on which the Condition depends

    Satisfied when:

        - the node specified in **dependency** executed in the previous `TIME_STEP`.

    Notes:

        - This Condition can transcend divisions between `TimeScales <TimeScale>`.
          For example, if A runs in the final `TIME_STEP` of an `TRIAL <TimeScale.TRIAL>`,
          JustRan(A) is satisfied at the beginning of the next `TRIAL <TimeScale.TRIAL>`.

    """
    def __init__(self, dependency) -> None: ...

class AllHaveRun(_DependencyValidation, Condition):

    """AllHaveRun

    Parameters:

        *dependencies (Hashable):  an iterable of nodes on which the Condition depends

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.TRIAL)

    Satisfied when:

        - all of the nodes specified in **dependencies** have executed at least once
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, *dependencies, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class WhenFinished(_DependencyValidation, Condition):

    """WhenFinished

    Parameters:

        dependency (Hashable):  the node on which the Condition depends

    Satisfied when:

        - the `is_finished` methods of the node specified in **dependencies** returns `True`.

    Notes:

        - This is a dynamic Condition: Each node is responsible for managing its finished status on its
          own, which can occur independently of the execution of other nodes.  Therefore the satisfaction of
          this Condition) can vary arbitrarily in time.

        - The is_finished method is called with `execution_id` as its           sole positional argument

    """
    def __init__(self, dependency) -> None: ...

class WhenFinishedAny(_DependencyValidation, Condition):

    """WhenFinishedAny

    Parameters:

        *dependencies (Hashable):  zero or more nodes on which the Condition depends

    Satisfied when:

        - the `is_finished` methods of any nodes specified in **dependencies** returns `True`.

    Notes:

        - This is a convenience class; WhenFinishedAny(A, B, C) is equivalent to
          Any(WhenFinished(A), WhenFinished(B), WhenFinished(C)).
          If no nodes are specified, the condition will default to checking all of scheduler's nodes.

        - This is a dynamic Condition: Each node is responsible for managing its finished status on its
          own, which can occur independently of the execution of other nodes.  Therefore the satisfaction of
          this Condition) can vary arbitrarily in time.

        - The is_finished method is called with `execution_id` as its           sole positional argument

    """
    def __init__(self, *dependencies) -> None: ...

class WhenFinishedAll(_DependencyValidation, Condition):

    """WhenFinishedAll

    Parameters:

        *dependencies (Hashable):  zero or more nodes on which the Condition depends

    Satisfied when:

        - the `is_finished` methods of all nodes specified in **dependencies** return `True`.

    Notes:

        - This is a convenience class; WhenFinishedAny(A, B, C) is equivalent to
          All(WhenFinished(A), WhenFinished(B), WhenFinished(C)).
          If no nodes are specified, the condition will default to checking all of scheduler's nodes.

        - This is a dynamic Condition: Each node is responsible for managing its finished status on its
          own, which can occur independently of the execution of other nodes.  Therefore the satisfaction of
          this Condition) can vary arbitrarily in time.

        - The is_finished method is called with `execution_id` as its           sole positional argument

    """
    def __init__(self, *dependencies) -> None: ...

class AtTrialStart(AtPass):

    """AtTrialStart

    Satisfied when:

        - at the beginning of an `TRIAL <TimeScale.TRIAL>`

    Notes:

        - identical to `AtPass(0) <AtPass>`
    """
    def __init__(self) -> None: ...

class AtTrialNStart(All):

    """AtTrialNStart

    Parameters:

        n(int): the `TRIAL <TimeScale.TRIAL>` on which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `TRIAL <TimeScale.TRIAL>`\\ s
        (default: TimeScale.RUN)

    Satisfied when:

        - on `PASS` 0 of the specified `TRIAL <TimeScale.TRIAL>` counted using 'TimeScale`

    Notes:

        - identical to All(AtPass(0), AtTrial(n, time_scale))

    """
    def __init__(self, n, time_scale: graph_scheduler.time.TimeScale = ...) -> None: ...

class AtRunStart(AtTrial):

    """AtRunStart

    Satisfied when:

        - at the beginning of an `RUN`

    Notes:

        - identical to `AtTrial(0) <AtTrial>`
    """
    def __init__(self) -> None: ...

class AtRunNStart(All):

    """AtRunNStart

    Parameters:

        n(int): the `RUN` on which the Condition is satisfied

    Satisfied when:

        - on `TRIAL <TimeScale.TRIAL>` 0 of the specified `RUN` counted using 'TimeScale`

    Notes:

        - identical to `All(AtTrial(0), AtRun(n))`

    """
    def __init__(self, n) -> None: ...

class Threshold(_DependencyValidation, Condition):

    """Threshold

    Attributes:

        dependency
            the node on which the Condition depends

        parameter
            the name of the parameter of **dependency** whose value is
            to be compared to **threshold**

        threshold
            the fixed value compared to the value of the **parameter**

        comparator
            the string comparison operator determining the direction or
            type of comparison of the value of the **parameter**
            relative to **threshold**

        indices
            if specified, a series of indices that reach the desired
            number given an iterable value for **parameter**

        atol
            absolute tolerance for the comparison

        rtol
            relative tolerance (to **threshold**) for the comparison

        custom_parameter_getter
            if specified, a function that returns the value of
            **parameter** for **dependency**; to support class
            structures other than <**dependency**>.<**parameter**>
            without subclassing

        custom_parameter_validator
            if specified, a function that throws an exception if there
            is no **parameter** for **dependency**; to support class
            structures other than <**dependency**>.<**parameter**>
            without subclassing

    Satisfied when:

        The comparison between the value of the **parameter** and
        **threshold** using **comparator** is true. If **comparator** is
        an equality (==, !=), the comparison will be considered equal
        within tolerances **atol** and **rtol**.

    Notes:

        The comparison must be done with scalars. If the value of
        **parameter** contains more than one item, **indices** must be
        specified.
    """
    def __init__(self, dependency, parameter, threshold, comparator, indices: Incomplete | None = ..., atol: int = ..., rtol: int = ..., custom_parameter_getter: Incomplete | None = ..., custom_parameter_validator: Incomplete | None = ...) -> None: ...
    def get_parameter_value(self, execution_id: Incomplete | None = ...): ...
    def validate_parameter(self, dependency, parameter, custom_parameter_validator: Incomplete | None = ...): ...

class GraphStructureCondition(ConditionBase):

    """
    Abstract base class for `graph structure conditions
    <Condition_Graph_Structure_Intro>`

    Subclasses must implement:
        `_process`
    """
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def modify_graph(self, graph: GraphDependencyDict) -> GraphDependencyDict:
        """
        Modifies **graph** based on the transformation specified by this
        condition

        Args:
            graph: a graph dependency dictionary

        Raises:
            ConditionError

        Returns:
            A copy of **graph** with modifications applied
        """
    def __init__(self, _owner: Hashable = ..., **kwargs) -> None: ...

class CustomGraphStructureCondition(GraphStructureCondition):

    """
    Applies a user-defined function to a graph

    Args:
        process_graph_function (Callable): a function taking an optional
            'self' argument (as the first argument, if present), and a
            graph dependency dictionary
        kwargs (**kwargs): optional arguments to be stored as attributes
    """
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, process_graph_function: Callable, **kwargs) -> None: ...

class _GSCUsingNodes(GraphStructureCondition):

    """
    Attributes:
        nodes: the subject nodes
    """
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, *nodes: Hashable, **kwargs) -> None: ...

class _GSCSingleNode(_GSCUsingNodes):

    """
    Attributes:
        node: the subject node
    """
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, node: Hashable, **kwargs) -> None: ...
    @property
    def node(self): ...

class _GSCWithOperations(_GSCUsingNodes):

    """
    Args:
        owner_senders: `Operation` that determines how the original
            senders of `owner <ConditionBase.owner>` (the Operation
            source) combine with the union of all original senders of
            all subject `nodes <_GSCUsingNodes.nodes>` (the
            Operation comparison) to produce the new set of senders of
            `owner <ConditionBase.owner>` after `modify_graph`
        owner_receivers: `Operation` that determines how the
            original receivers of `owner <ConditionBase.owner>` (the
            Operation source) combine with the union of all original
            receivers of all subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation comparison)
            to produce the new set of receivers of `owner
            <ConditionBase.owner>` after `modify_graph`
        subject_senders: `Operation` that determines how the
            original senders for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original senders of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of senders for
            the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        subject_receivers: `Operation` that determines how the
            original receivers for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original receivers of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of receivers
            for the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        reconnect_non_subject_receivers: If True, `modify_graph`
            will create an edge from all prior senders of `owner` to
            all receivers of `owner` that are not in `nodes`, if
            there is no longer a path from that sender to that
            receiver.
            Defaults to True.
        remove_new_self_referential_edges: If True, `modify_graph`
            will remove any newly-created edges from a node to
            itself.
            Defaults to True.
        prune_cycles: If True, `modify_graph` will attempt to prune
            any newly-created cycles, preferring to remove edges
            adjacent to `owner` that affect the placement of `owner`
            more than any subject `node <nodes>`.
            Defaults to True.
        ignore_conflicts: If True, when any two operations give
            different results for the new senders and receivers of a
            node in `modify_graph`, an error will not be raised.
            Defaults to False.

    Attributes:
        nodes: the subject nodes
    """
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, *nodes: Hashable, owner_senders: Operation | str = ..., owner_receivers: Operation | str = ..., subject_senders: SubjectOperation = ..., subject_receivers: SubjectOperation = ..., reconnect_non_subject_receivers: bool = ..., remove_new_self_referential_edges: bool = ..., prune_cycles: bool = ..., ignore_conflicts: bool = ..., **kwargs) -> None: ...

class _GSCReposition(_GSCUsingNodes):
    _already_valid_message: ClassVar[str] = ...
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, *nodes: Hashable, **kwargs) -> None: ...

class BeforeNodes(_GSCReposition, _GSCWithOperations):

    """
    Adds a dependency from the owner to each of the specified nodes and
    optionally modifies the senders and receivers of all affected nodes

    Args:
        owner_senders: `Operation` that determines how the original
            senders of `owner <ConditionBase.owner>` (the Operation
            source) combine with the union of all original senders of
            all subject `nodes <_GSCUsingNodes.nodes>` (the
            Operation comparison) to produce the new set of senders of
            `owner <ConditionBase.owner>` after `modify_graph`
        owner_receivers: `Operation` that determines how the
            original receivers of `owner <ConditionBase.owner>` (the
            Operation source) combine with the union of all original
            receivers of all subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation comparison)
            to produce the new set of receivers of `owner
            <ConditionBase.owner>` after `modify_graph`
        subject_senders: `Operation` that determines how the
            original senders for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original senders of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of senders for
            the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        subject_receivers: `Operation` that determines how the
            original receivers for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original receivers of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of receivers
            for the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        reconnect_non_subject_receivers: If True, `modify_graph`
            will create an edge from all prior senders of `owner` to
            all receivers of `owner` that are not in `nodes`, if
            there is no longer a path from that sender to that
            receiver.
            Defaults to True.
        remove_new_self_referential_edges: If True, `modify_graph`
            will remove any newly-created edges from a node to
            itself.
            Defaults to True.
        prune_cycles: If True, `modify_graph` will attempt to prune
            any newly-created cycles, preferring to remove edges
            adjacent to `owner` that affect the placement of `owner`
            more than any subject `node <nodes>`.
            Defaults to True.
        ignore_conflicts: If True, when any two operations give
            different results for the new senders and receivers of a
            node in `modify_graph`, an error will not be raised.
            Defaults to False.

    Attributes:
        nodes: the subject nodes
    """
    _already_valid_message: ClassVar[str] = ...
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, *nodes, owner_senders: Operation | str = ..., owner_receivers: Operation | str = ..., subject_senders: SubjectOperation = ..., subject_receivers: SubjectOperation = ..., reconnect_non_subject_receivers: bool = ..., remove_new_self_referential_edges: bool = ..., prune_cycles: bool = ..., ignore_conflicts: bool = ...) -> None: ...

class BeforeNode(BeforeNodes, _GSCSingleNode):

    """
    Adds a dependency from the owner to the specified node and
    optionally modifies the senders and receivers of both

    Args:
        owner_senders: `Operation` that determines how the original
            senders of `owner <ConditionBase.owner>` (the Operation
            source) combine with the union of all original senders of
            all subject `nodes <_GSCUsingNodes.nodes>` (the
            Operation comparison) to produce the new set of senders of
            `owner <ConditionBase.owner>` after `modify_graph`
        owner_receivers: `Operation` that determines how the
            original receivers of `owner <ConditionBase.owner>` (the
            Operation source) combine with the union of all original
            receivers of all subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation comparison)
            to produce the new set of receivers of `owner
            <ConditionBase.owner>` after `modify_graph`
        subject_senders: `Operation` that determines how the
            original senders for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original senders of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of senders for
            the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        subject_receivers: `Operation` that determines how the
            original receivers for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original receivers of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of receivers
            for the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        reconnect_non_subject_receivers: If True, `modify_graph`
            will create an edge from all prior senders of `owner` to
            all receivers of `owner` that are not in `nodes`, if
            there is no longer a path from that sender to that
            receiver.
            Defaults to True.
        remove_new_self_referential_edges: If True, `modify_graph`
            will remove any newly-created edges from a node to
            itself.
            Defaults to True.
        prune_cycles: If True, `modify_graph` will attempt to prune
            any newly-created cycles, preferring to remove edges
            adjacent to `owner` that affect the placement of `owner`
            more than any subject `node <nodes>`.
            Defaults to True.
        ignore_conflicts: If True, when any two operations give
            different results for the new senders and receivers of a
            node in `modify_graph`, an error will not be raised.
            Defaults to False.

    Attributes:
        nodes: the subject nodes

    Attributes:
        node: the subject node
    """
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, *nodes, owner_senders: Operation | str = ..., owner_receivers: Operation | str = ..., subject_senders: SubjectOperation = ..., subject_receivers: SubjectOperation = ..., reconnect_non_subject_receivers: bool = ..., remove_new_self_referential_edges: bool = ..., prune_cycles: bool = ..., ignore_conflicts: bool = ...) -> None: ...

class WithNode(_GSCReposition, _GSCWithOperations, _GSCSingleNode):

    """
    Adds a dependency from each of the senders of both the owner and the
    specified node to both the owner and the specified node, and
    optionally modifies the receivers of both

    Args:
        owner_senders: `Operation` that determines how the original
            senders of `owner <ConditionBase.owner>` (the Operation
            source) combine with the union of all original senders of
            all subject `nodes <_GSCUsingNodes.nodes>` (the
            Operation comparison) to produce the new set of senders of
            `owner <ConditionBase.owner>` after `modify_graph`
        owner_receivers: `Operation` that determines how the
            original receivers of `owner <ConditionBase.owner>` (the
            Operation source) combine with the union of all original
            receivers of all subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation comparison)
            to produce the new set of receivers of `owner
            <ConditionBase.owner>` after `modify_graph`
        subject_senders: `Operation` that determines how the
            original senders for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original senders of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of senders for
            the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        subject_receivers: `Operation` that determines how the
            original receivers for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original receivers of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of receivers
            for the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        reconnect_non_subject_receivers: If True, `modify_graph`
            will create an edge from all prior senders of `owner` to
            all receivers of `owner` that are not in `nodes`, if
            there is no longer a path from that sender to that
            receiver.
            Defaults to True.
        remove_new_self_referential_edges: If True, `modify_graph`
            will remove any newly-created edges from a node to
            itself.
            Defaults to True.
        prune_cycles: If True, `modify_graph` will attempt to prune
            any newly-created cycles, preferring to remove edges
            adjacent to `owner` that affect the placement of `owner`
            more than any subject `node <nodes>`.
            Defaults to True.
        ignore_conflicts: If True, when any two operations give
            different results for the new senders and receivers of a
            node in `modify_graph`, an error will not be raised.
            Defaults to False.

    Attributes:
        nodes: the subject nodes

    Attributes:
        node: the subject node
    """
    _already_valid_message: ClassVar[str] = ...
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, node, owner_receivers: Operation | str = ..., subject_receivers: SubjectOperation = ..., reconnect_non_subject_receivers: bool = ..., remove_new_self_referential_edges: bool = ..., prune_cycles: bool = ..., ignore_conflicts: bool = ...) -> None: ...

class AfterNodes(_GSCReposition, _GSCWithOperations):

    """
    Adds a dependency from each of the specified nodes to the owner
    and optionally modifies the senders and receivers of all
    affected nodes

    Args:
        owner_senders: `Operation` that determines how the original
            senders of `owner <ConditionBase.owner>` (the Operation
            source) combine with the union of all original senders of
            all subject `nodes <_GSCUsingNodes.nodes>` (the
            Operation comparison) to produce the new set of senders of
            `owner <ConditionBase.owner>` after `modify_graph`
        owner_receivers: `Operation` that determines how the
            original receivers of `owner <ConditionBase.owner>` (the
            Operation source) combine with the union of all original
            receivers of all subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation comparison)
            to produce the new set of receivers of `owner
            <ConditionBase.owner>` after `modify_graph`
        subject_senders: `Operation` that determines how the
            original senders for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original senders of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of senders for
            the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        subject_receivers: `Operation` that determines how the
            original receivers for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original receivers of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of receivers
            for the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        reconnect_non_subject_receivers: If True, `modify_graph`
            will create an edge from all prior senders of `owner` to
            all receivers of `owner` that are not in `nodes`, if
            there is no longer a path from that sender to that
            receiver.
            Defaults to True.
        remove_new_self_referential_edges: If True, `modify_graph`
            will remove any newly-created edges from a node to
            itself.
            Defaults to True.
        prune_cycles: If True, `modify_graph` will attempt to prune
            any newly-created cycles, preferring to remove edges
            adjacent to `owner` that affect the placement of `owner`
            more than any subject `node <nodes>`.
            Defaults to True.
        ignore_conflicts: If True, when any two operations give
            different results for the new senders and receivers of a
            node in `modify_graph`, an error will not be raised.
            Defaults to False.

    Attributes:
        nodes: the subject nodes
    """
    _already_valid_message: ClassVar[str] = ...
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, *nodes, owner_senders: Operation | str = ..., owner_receivers: Operation | str = ..., subject_senders: SubjectOperation = ..., subject_receivers: SubjectOperation = ..., reconnect_non_subject_receivers: bool = ..., remove_new_self_referential_edges: bool = ..., prune_cycles: bool = ..., ignore_conflicts: bool = ...) -> None: ...

class AfterNode(AfterNodes, _GSCSingleNode):

    """
    Adds a dependency from the specified node to the owner and
    optionally modifies the senders and receivers of both

    Args:
        owner_senders: `Operation` that determines how the original
            senders of `owner <ConditionBase.owner>` (the Operation
            source) combine with the union of all original senders of
            all subject `nodes <_GSCUsingNodes.nodes>` (the
            Operation comparison) to produce the new set of senders of
            `owner <ConditionBase.owner>` after `modify_graph`
        owner_receivers: `Operation` that determines how the
            original receivers of `owner <ConditionBase.owner>` (the
            Operation source) combine with the union of all original
            receivers of all subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation comparison)
            to produce the new set of receivers of `owner
            <ConditionBase.owner>` after `modify_graph`
        subject_senders: `Operation` that determines how the
            original senders for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original senders of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of senders for
            the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        subject_receivers: `Operation` that determines how the
            original receivers for each of the subject `nodes
            <_GSCUsingNodes.nodes>` (the Operation source) combine with
            the original receivers of `owner <ConditionBase.owner>` (the
            Operation comparison) to produce the new set of receivers
            for the subject `nodes <_GSCUsingNodes.nodes>` after
            `modify_graph`. Operations are applied individually to each
            subject node, and this argument may also be specified as a
            dictionary mapping nodes to separate operations.
        reconnect_non_subject_receivers: If True, `modify_graph`
            will create an edge from all prior senders of `owner` to
            all receivers of `owner` that are not in `nodes`, if
            there is no longer a path from that sender to that
            receiver.
            Defaults to True.
        remove_new_self_referential_edges: If True, `modify_graph`
            will remove any newly-created edges from a node to
            itself.
            Defaults to True.
        prune_cycles: If True, `modify_graph` will attempt to prune
            any newly-created cycles, preferring to remove edges
            adjacent to `owner` that affect the placement of `owner`
            more than any subject `node <nodes>`.
            Defaults to True.
        ignore_conflicts: If True, when any two operations give
            different results for the new senders and receivers of a
            node in `modify_graph`, an error will not be raised.
            Defaults to False.

    Attributes:
        nodes: the subject nodes

    Attributes:
        node: the subject node
    """
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, *nodes, owner_senders: Operation | str = ..., owner_receivers: Operation | str = ..., subject_senders: SubjectOperation = ..., subject_receivers: SubjectOperation = ..., reconnect_non_subject_receivers: bool = ..., remove_new_self_referential_edges: bool = ..., prune_cycles: bool = ..., ignore_conflicts: bool = ...) -> None: ...

class AddEdgeTo(_GSCSingleNode):

    """
    Adds an edge from `AddEdgeTo.owner <ConditionBase.owner>` to
    `AddEdgeTo.node`

    Attributes:
        node: the subject node
    """
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, node: Hashable) -> None: ...

class RemoveEdgeFrom(_GSCSingleNode):

    """
    Removes an edge from `RemoveEdgeFrom.node` to `RemoveEdgeFrom.owner
    <ConditionBase.owner>`

    Attributes:
        node: the subject node
    """
    __abstractmethods__: ClassVar[frozenset] = ...
    _abc_impl: ClassVar[_abc._abc_data] = ...
    def __init__(self, node: Hashable) -> None: ...



from psyneulink.core.llvm import ExecutionMode
from psyneulink.core.globals.keywords import ALL, ADAPTIVE, CONTROL, CPU, Loss, MPS, OPTIMIZATION_STEP, RUN, TRIAL



model_params = dict(

    # Names:
    name = "EGO Model CSW",
    em_name = "EM",
    state_input_layer_name = "STATE",
    previous_state_layer_name = "PREVIOUS STATE",
    context_layer_name = 'CONTEXT',
    prediction_layer_name = "PREDICTION",

    # Structural
    state_d = 11, # length of state vector
    previous_state_d = 11, # length of state vector
    context_d = 11, # length of context vector
    memory_capacity = ALL, # number of entries in EM memory; ALL=> match to number of stims
    memory_init = (0,.0001),  # Initialize memory with random values in interval
    # memory_init = None,  # Initialize with zeros
    concatenate_queries = False,
    # concatenate_queries = True,
    # environment
    # curriculum_type = 'Interleaved',
    curriculum_type = 'Blocked',
    num_stims = 7,  # Integer or ALL
    # num_stims = ALL,  # Integer or ALL

    # Processing
    integration_rate = .69, # rate at which state is integrated into new context
    state_weight = 1, # weight of the state used during memory retrieval
    context_weight = 1, # weight of the context used during memory retrieval
    # normalize_field_weights = False, # whether to normalize the field weights during memory retrieval
    normalize_field_weights = True, # whether to normalize the field weights during memory retrieval
    normalize_memories = False, # whether to normalize the memory during memory retrieval
    # softmax_temperature = None, # temperature of the softmax used during memory retrieval (smaller means more argmax-like
    softmax_temperature = .1, # temperature of the softmax used during memory retrieval (smaller means more argmax-like
    # softmax_temperature = ADAPTIVE, # temperature of the softmax used during memory retrieval (smaller means more argmax-like
    # softmax_temperature = CONTROL, # temperature of the softmax used during memory retrieval (smaller means more argmax-like
    # softmax_threshold = None, # threshold used to mask out small values in softmax
    softmax_threshold = .001, # threshold used to mask out small values in softmax
    # target_fields=[True, False, False], # Enable learning for PREDICTION (STATE) but not CONTEXT or PREVIOUS STATE
    # target_fields=[True, True, True]
    # target_fields=True,
    # target_fields=False,
    enable_learning = True,
    # enable_learning = False,
    loss_spec = Loss.BINARY_CROSS_ENTROPY,
    # loss_spec = Loss.CROSS_ENTROPY,
    # loss_spec = Loss.MSE,
    learning_rate = .5,
    num_optimization_steps = 10,
    synch_weights = RUN,
    synch_values = RUN,
    synch_results = RUN,
    # execution_mode = ExecutionMode.Python,
    execution_mode = ExecutionMode.PyTorch,
    device = CPU,
    # device = MPS,
)
#endregion


# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* Scheduler **************************************************************
import copy
import logging
import typing
from typing import Hashable

import graph_scheduler
import pint

import psyneulink as pnl
from psyneulink import _unit_registry
from psyneulink.core.globals.context import Context, handle_external_context
from psyneulink.core.globals.mdf import MDFSerializable
from psyneulink.core.globals.utilities import parse_valid_identifier
from psyneulink.core.scheduling.condition import _create_as_pnl_condition, graph_structure_conditions_available, gsc_unavailable_message

__all__ = [
    'Scheduler', 'SchedulingMode'
]


logger = logging.getLogger(__name__)
SchedulingMode = graph_scheduler.scheduler.SchedulingMode


class Scheduler(graph_scheduler.Scheduler, MDFSerializable):
    def __init__(
        self,
        composition=None,
        graph=None,
        conditions=None,
        termination_conds=None,
        default_execution_id=None,
        mode: SchedulingMode = SchedulingMode.STANDARD,
        default_absolute_time_unit: typing.Union[str, pint.Quantity] = 1 * _unit_registry.ms,
        **kwargs
    ):
        """
        :param composition: (Composition) - the Composition this scheduler is scheduling for
        """

        if composition is not None:
            # dependency dict
            graph = composition.graph_processing.prune_feedback_edges()[0]
            if default_execution_id is None:
                default_execution_id = composition.default_execution_id
            self.composition = composition

        # TODO: consider integrating something like this into graph-scheduler?
        self._user_specified_conds = graph_scheduler.ConditionSet()
        self._user_specified_termination_conds = copy.copy(termination_conds) if termination_conds is not None else {}

        super().__init__(
            graph=graph,
            conditions=conditions,
            termination_conds=termination_conds,
            default_execution_id=default_execution_id,
            mode=mode,
            default_absolute_time_unit=default_absolute_time_unit,
            **kwargs,
        )

        def replace_term_conds(term_conds):
            return {
                ts: _create_as_pnl_condition(cond) for ts, cond in term_conds.items()
            }

        self.default_termination_conds = replace_term_conds(self.default_termination_conds)
        self.termination_conds = replace_term_conds(self.termination_conds)

    def _validate_conditions(self):
        unspecified_nodes = []

        # pre-graph-structure-condition compatibility
        try:
            conditions_basic = self.conditions.conditions_basic
        except AttributeError:
            conditions_basic = self.conditions.conditions

        for node in self.nodes:
            if node not in conditions_basic:
                dependencies = list(self.dependency_dict[node])
                if len(dependencies) == 0:
                    cond = graph_scheduler.Always()
                elif len(dependencies) == 1:
                    cond = graph_scheduler.EveryNCalls(dependencies[0], 1)
                else:
                    cond = graph_scheduler.All(*[graph_scheduler.EveryNCalls(x, 1) for x in dependencies])

                # TODO: replace this call in graph-scheduler if adding _user_specified_conds
                self._add_condition(node, cond)
                unspecified_nodes.append(node)

        super()._validate_conditions()

    def add_condition(self, owner, condition):
        self._user_specified_conds.add_condition(owner, condition)
        self._add_condition(owner, condition)

    def _add_condition(self, owner, condition):
        condition = _create_as_pnl_condition(condition)
        super().add_condition(owner, condition)

        if graph_structure_conditions_available:
            if isinstance(condition, pnl.GraphStructureCondition):
                self.composition._analyze_graph()

    def add_condition_set(self, conditions):
        self._user_specified_conds.add_condition_set(conditions)
        self._add_condition_set(conditions)

    def _add_condition_set(self, conditions):
        conditions = {
            node: _create_as_pnl_condition(conditions[node])
            for node in conditions
        }
        super().add_condition_set(conditions)

    def remove_condition(self, owner_or_condition):
        try:
            res = super().remove_condition(owner_or_condition)
        except AttributeError as e:
            if "has no attribute 'remove_condition'" in str(e):
                raise graph_scheduler.SchedulerError(
                    f'remove_condition unavailable in your installed graph-scheduler v{graph_scheduler.__version__}'
                )
            else:
                raise
        else:
            if isinstance(res, pnl.GraphStructureCondition):
                self.composition._analyze_graph()

        return res

    @graph_scheduler.Scheduler.termination_conds.setter
    def termination_conds(self, termination_conds):
        if termination_conds is not None:
            self._user_specified_termination_conds.update(termination_conds)

        graph_scheduler.Scheduler.termination_conds.fset(self, termination_conds)

    @handle_external_context(fallback_default=True)
    def run(
        self,
        termination_conds=None,
        context=None,
        base_context=Context(execution_id=None),
        skip_trial_time_increment=False,
    ):
        yield from super().run(
            termination_conds=termination_conds,
            context=context,
            execution_id=context.execution_id,
            base_execution_id=base_context.execution_id,
            skip_environment_state_update_time_increment=skip_trial_time_increment,
        )

    def as_mdf_model(self):
        import modeci_mdf.mdf as mdf

        return mdf.ConditionSet(
            node_specific={
                parse_valid_identifier(n.name): self.conditions[n].as_mdf_model() for n in self.nodes if n in self.conditions
            },
            termination={
                str.lower(k.name): v.as_mdf_model() for k, v in self.termination_conds.items()
            },
        )

    @handle_external_context()
    def get_clock(self, context):
        return super().get_clock(context.execution_id)

    def add_graph_edge(self, sender: Hashable, receiver: Hashable) -> 'pnl.AddEdgeTo':
        if not graph_structure_conditions_available:
            raise graph_scheduler.SchedulerError(gsc_unavailable_message)

        cond = pnl.AddEdgeTo(receiver)
        self.add_condition(sender, cond)
        return cond

    def remove_graph_edge(self, sender: Hashable, receiver: Hashable) -> 'pnl.RemoveEdgeFrom':
        if not graph_structure_conditions_available:
            raise graph_scheduler.SchedulerError(gsc_unavailable_message)

        cond = pnl.RemoveEdgeFrom(sender)
        self.add_condition(receiver, cond)
        return cond


_doc_subs = {
    None: [
        (
            '(When creating a Scheduler explicitly, the set of nodes)',
            'A Scheduler can be created explicitly using its constructor.  However, more commonly it is created automatically for a `Composition <Composition>` when it is created.\\1',
        ),

        (
            r'(\n\* a \*graph specification dictionary\* in the \*\*graph\*\* argument)',
            "\n* a `Composition` in the **composition** argument - if a Composition is specified, the Scheduler is created using the nodes and edges in the Composition's `graph <Composition.graph_processing>`, with any `feedback Projections <Composition_Feedback>` pruned as needed to produce an acyclic graph. If there is a cycle comprised of all non-feedback projections, this cycle is reduced to a single `consideration set <consideration_set>`\n\\1",
        ),

        ('origin nodes', '`ORIGIN` nodes',),

        (
            r'Examples\n--------.*\n\.\. _Scheduler_Class_Reference',
            """
Examples
--------

Please see `Condition` for a list of all supported Conditions and their behavior.

* Basic phasing in a linear process::

    >>> import psyneulink as pnl

    >>> A = pnl.TransferMechanism(name='A')
    >>> B = pnl.TransferMechanism(name='B')
    >>> C = pnl.TransferMechanism(name='C')

    >>> comp = pnl.Composition()

    >>> pway = comp.add_linear_processing_pathway([A, B, C])
    >>> pway.pathway
    [(TransferMechanism A), (MappingProjection MappingProjection from A[RESULT] to B[InputPort-0]), (TransferMechanism B), (MappingProjection MappingProjection from B[RESULT] to C[InputPort-0]), (TransferMechanism C)]

    >>> # implicit condition of Always for A
    >>> comp.scheduler.add_condition(B, pnl.EveryNCalls(A, 2))
    >>> comp.scheduler.add_condition(C, pnl.EveryNCalls(B, 3))

    >>> # implicit AllHaveRun Termination condition
    >>> execution_sequence = list(comp.scheduler.run())
    >>> execution_sequence
    [{(TransferMechanism A)}, {(TransferMechanism A)}, {(TransferMechanism B)}, {(TransferMechanism A)}, {(TransferMechanism A)}, {(TransferMechanism B)}, {(TransferMechanism A)}, {(TransferMechanism A)}, {(TransferMechanism B)}, {(TransferMechanism C)}]

* Alternate basic phasing in a linear process::

    >>> comp = pnl.Composition()
    >>> pway = comp.add_linear_processing_pathway([A, B])
    >>> pway.pathway
    [(TransferMechanism A), (MappingProjection MappingProjection from A[RESULT] to B[InputPort-0]), (TransferMechanism B)]

    >>> comp.scheduler.add_condition(
    ...     A,
    ...     pnl.Any(
    ...         pnl.AtPass(0),
    ...         pnl.EveryNCalls(B, 2)
    ...     )
    ... )

    >>> comp.scheduler.add_condition(
    ...     B,
    ...     pnl.Any(
    ...         pnl.EveryNCalls(A, 1),
    ...         pnl.EveryNCalls(B, 1)
    ...     )
    ... )
    >>> termination_conds = {
    ...     pnl.TimeScale.TRIAL: pnl.AfterNCalls(B, 4, time_scale=pnl.TimeScale.TRIAL)
    ... }
    >>> execution_sequence = list(comp.scheduler.run(termination_conds=termination_conds))
    >>> execution_sequence # doctest: +SKIP
    [{(TransferMechanism A)}, {(TransferMechanism B)}, {(TransferMechanism B)}, {(TransferMechanism A)}, {(TransferMechanism B)}, {(TransferMechanism B)}]

* Basic phasing in two processes::

    >>> comp = pnl.Composition()
    >>> pway = comp.add_linear_processing_pathway([A, C])
    >>> pway.pathway
    [(TransferMechanism A), (MappingProjection MappingProjection from A[RESULT] to C[InputPort-0]), (TransferMechanism C)]

    >>> pway = comp.add_linear_processing_pathway([B, C])
    >>> pway.pathway
    [(TransferMechanism B), (MappingProjection MappingProjection from B[RESULT] to C[InputPort-0]), (TransferMechanism C)]

    >>> comp.scheduler.add_condition(A, pnl.EveryNPasses(1))
    >>> comp.scheduler.add_condition(B, pnl.EveryNCalls(A, 2))
    >>> comp.scheduler.add_condition(
    ...     C,
    ...     pnl.Any(
    ...         pnl.AfterNCalls(A, 3),
    ...         pnl.AfterNCalls(B, 3)
    ...     )
    ... )
    >>> termination_conds = {
    ...     pnl.TimeScale.TRIAL: pnl.AfterNCalls(C, 4, time_scale=pnl.TimeScale.TRIAL)
    ... }
    >>> execution_sequence = list(comp.scheduler.run(termination_conds=termination_conds))
    >>> execution_sequence  # doctest: +SKIP
    [{(TransferMechanism A)}, {(TransferMechanism A), (TransferMechanism B)}, {(TransferMechanism A)}, {(TransferMechanism C)}, {(TransferMechanism A), (TransferMechanism B)}, {(TransferMechanism C)}, {(TransferMechanism A)}, {(TransferMechanism C)}, {(TransferMechanism A), (TransferMechanism B)}, {(TransferMechanism C)}]

.. _Scheduler_Class_Reference
"""
        ),

        (
            r'linear graph with two nodes.*In standard mode',
            """linear composition with two nodes::

    >>> import psyneulink as pnl

    >>> A = pnl.TransferMechanism()
    >>> B = pnl.TransferMechanism()

    >>> comp = pnl.Composition()
    >>> pway = comp.add_linear_processing_pathway([A, B])

    >>> comp.scheduler.add_condition(A, pnl.TimeInterval(start=10))
    >>> comp.scheduler.add_condition(B, pnl.TimeInterval(start=10))

In standard mode"""
        ),

        (
            r'(earlier or later \(e\.g\., when the Condition\nfor a particular node or set of nodes is met\)\.)',
            '\\1 When the Scheduler terminates a `TRIAL <TimeScale.TRIAL>`, the `Composition <Composition>` begins processing the next input specified in the call to its `run <Composition.run>` method. Thus, a `TRIAL <TimeScale.TRIAL>` is defined as the scope of processing associated with a given input to the Composition.'
        ),

        (
            '(is when all of its constituent environment state updates have terminated.)',
            """\\1 These defaults may be overriden when running a Composition, by passing a dictionary mapping `TimeScales <TimeScale>` to `Conditions <Condition>` in the **termination_processing** argument of a call to `Composition.run` (to terminate the execution of processing)::

    Composition.run(
        ...,
        termination_processing={TimeScale.TRIAL: WhenFinished(ddm)}
        )
"""
        ),

        (
            r'(however, using the `indices in the original consideration queue<.*?\.Scheduler\.consideration_queue_indices>`\.)',
            '\\1 `Composition`\\ s will execute nodes in this order, however independent usages of the scheduler may not. The user should be aware of this and set up defaults and inputs to nodes accordingly.'
        ),
    ],
    'Scheduler': [
        (
            r'(Arguments\n    ---------\n)',
            """\\1
    composition : Composition
        specifies the `Components <Component>` to be ordered for execution, and any dependencies among them,
        based on the `Composition <Composition>`\\'s `graph <Composition.graph_processing>`.
"""
        )
    ]
}



import numpy as np
import psyneulink as pnl

# Mechanisms
Input = pnl.TransferMechanism(name='Input')
reward = pnl.TransferMechanism(
    output_ports=[pnl.RESULT, pnl.MEAN, pnl.VARIANCE], name='reward'
)
Decision = pnl.DDM(
    function=pnl.DriftDiffusionAnalytical(
        drift_rate=(
            1.0,
            pnl.ControlProjection(
                function=pnl.Linear,
                control_signal_params={
                    pnl.ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)
                },
            ),
        ),
        threshold=(
            1.0,
            pnl.ControlProjection(
                function=pnl.Linear,
                control_signal_params={
                    pnl.ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)
                },
            ),
        ),
        noise=0.5,
        starting_value=0,
        non_decision_time=0.45,
    ),
    output_ports=[
        pnl.DECISION_VARIABLE,
        pnl.RESPONSE_TIME,
        pnl.PROBABILITY_UPPER_THRESHOLD,
    ],
    name='Decision',
)

comp = pnl.Composition(name='comp')
comp.add_node(reward, required_roles=[pnl.NodeRole.OUTPUT])
comp.add_node(Decision, required_roles=[pnl.NodeRole.OUTPUT])
task_execution_pathway = [Input, pnl.IDENTITY_MATRIX, Decision]
comp.add_linear_processing_pathway(task_execution_pathway)

comp.add_controller(
    controller=pnl.OptimizationControlMechanism(
        agent_rep=comp,
        state_features=[Input.input_port, reward.input_port],
        state_feature_function=pnl.AdaptiveIntegrator(rate=0.5),
        objective_mechanism=pnl.ObjectiveMechanism(
            function=pnl.LinearCombination(operation=pnl.PRODUCT),
            monitor=[
                reward,
                Decision.output_ports[pnl.PROBABILITY_UPPER_THRESHOLD],
                (Decision.output_ports[pnl.RESPONSE_TIME], -1, 1),
            ],
        ),
        function=pnl.GridSearch,
        control=[
            {
                pnl.PROJECTIONS: ('drift_rate', Decision),
                pnl.ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3),
            },
            {
                pnl.PROJECTIONS: (pnl.THRESHOLD, Decision),
                pnl.ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3),
            },
        ],
    )
)



from sys import version_info

if version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

from beartype.typing import (
    Union as Union,
    Type as Type,
    Callable as Callable,
    Optional as Optional,
    Any as Any,
    Tuple as Tuple,
    List as List,
    Dict as Dict,
    Iterable as Iterable,
    Set as Set,
    Mapping as Mapping,
)



"""
bistable percepts
"""

import itertools
import numpy as np
import psyneulink as pnl
import matplotlib
matplotlib.use('TkAgg')
import matplotlib.pyplot as plt

# constants
# this code only works for N_PERCEPTS == 2
ALL_PERCEPTS = ['a', 'b']
N_PERCEPTS = len(ALL_PERCEPTS)
assert N_PERCEPTS == 2

# variables
n_nodes_per_percepts = 8
node_dict = {percept: None for percept in ALL_PERCEPTS}

excit_level = 1
inhib_level = 1


def get_node(percept, node_id):
    # helper func for creating a node
    tm_function = pnl.Linear(slope=1, intercept=0)
    tm_integrator_mode = True
    tm_integration_rate = .5
    node_ = pnl.TransferMechanism(
        name=f'{percept}-{node_id}',
        function=tm_function,
        integrator_mode=tm_integrator_mode,
        integration_rate=tm_integration_rate,
        default_variable=np.zeros((1,)),
    )
    return node_


# init all nodes, save them in list and dict form
for percept in ALL_PERCEPTS:
    node_dict[percept] = [
        get_node(percept, i) for i in range(n_nodes_per_percepts)
    ]


# init composition
bp_comp = pnl.Composition()

print('Forming connetions: ')
# within-percept excitation
for percept in ALL_PERCEPTS:
    for node_i, node_j in itertools.product(node_dict[percept], node_dict[percept]):
        if node_i is not node_j:
            print(f'\t{node_i} -> excite -> {node_j}')
            bp_comp.add_linear_processing_pathway(
                pathway=(node_i, [excit_level], node_j))

# inter-percepts inhibition
for node_i, node_j in zip(node_dict[ALL_PERCEPTS[0]],
                          node_dict[ALL_PERCEPTS[1]]):
    print(f'\t{node_i} <- inhibit -> {node_j}')
    bp_comp.add_linear_processing_pathway(
        pathway=(node_i, [-inhib_level], node_j))
    bp_comp.add_linear_processing_pathway(
        pathway=(node_j, [-inhib_level], node_i))

# make sure all nodes are both input and outputs
reportOutputPref = False
for node in bp_comp.nodes:
    # # MODIFIED 4/25/20 OLD:
    # bp_comp.add_required_node_role(node, pnl.NodeRole.INPUT)
    # bp_comp.add_required_node_role(node, pnl.NodeRole.OUTPUT)
    # MODIFIED 4/25/20 NEW:
    bp_comp.require_node_roles(node, [pnl.NodeRole.INPUT, pnl.NodeRole.OUTPUT])
    # MODIFIED 4/25/20 END
    # turn off report
    node.reportOutputPref = reportOutputPref

# bp_comp.show_graph()

# init the inputs
n_time_steps = 10
input_dict = {
    node_: np.random.normal(size=(n_time_steps,))
    for node_ in bp_comp.nodes
}

# run the model
bp_comp.run(input_dict, num_trials=10)

acts = np.squeeze(bp_comp.results)
f, ax = plt.subplots(1, 1, figsize=(8, 5))
ax.plot(acts[:, :n_nodes_per_percepts], color='red')
ax.plot(acts[:, n_nodes_per_percepts:], color='blue')
ax.set_xlabel('Time')
ax.set_ylabel('Activation')
ax.set_title('temporal dyanmics of the bistable percept model')



import psyneulink as pnl
from psyneulink.core.components.functions.nonstateful.transferfunctions import Logistic
import numpy as np

print(pnl.__version__)


# network params
n_input = 2
n_hidden = 5
n_output = 1
key_size = n_hidden
val_size = n_hidden
max_entries = 7

# training params
num_epochs = 3
learning_rate = .1
wts_init_scale = .1

# layers

input = pnl.TransferMechanism(
    name='input',
    default_variable=np.zeros(n_input)
)

hidden = pnl.TransferMechanism(
    name='hidden',
    default_variable=np.zeros(n_hidden),
    function=Logistic()
)

output = pnl.TransferMechanism(
    name='output',
    default_variable=np.zeros(n_output),
    function=Logistic()
)

# weights
w_ih = pnl.MappingProjection(
    name='input_to_hidden',
    matrix=np.random.randn(n_input, n_hidden) * wts_init_scale,
    sender=input,
    receiver=hidden
)

w_ho = pnl.MappingProjection(
    name='hidden_to_output',
    matrix=np.random.randn(n_hidden, n_output) * wts_init_scale,
    sender=hidden,
    receiver=output
)


# DictionaryMemory
EM = pnl.EpisodicMemoryMechanism(
    key_size, val_size,
    name='episodic memory'
)

w_hd = pnl.MappingProjection(
    name='hidden_to_em',
    matrix=np.random.randn(n_hidden, n_hidden) * wts_init_scale,
    sender=hidden,
    receiver=EM
)

w_dh = pnl.MappingProjection(
    name='em_to_hidden',
    matrix=np.random.randn(n_hidden, n_hidden) * wts_init_scale,
    sender=EM,
    receiver=hidden
)

comp = pnl.Composition(name='xor')
# add all nodes
all_nodes = [input, hidden, output, EM]
for node in all_nodes:
    comp.add_node(node)
# input-hidden-output pathway
comp.add_projection(sender=input, projection=w_ih, receiver=hidden)
comp.add_projection(sender=hidden, projection=w_ho, receiver=output)
# conneciton, DictionaryMemory
comp.add_projection(sender=hidden, projection=w_hd, receiver=EM)
comp.add_projection(sender=EM, projection=w_dh, receiver=hidden)

# comp.show_graph()

print(EM.input_ports.names)
print(EM.input_ports[pnl.CUE_INPUT].path_afferents[0].sender.owner.name)
print(hidden.input_ports[0].path_afferents[0].sender.owner.name)



from . import composition
from . import compositionfunctionapproximator
from . import parameterestimationcomposition
from . import showgraph


from .composition import *
from .pathway import *
from .compositionfunctionapproximator import *
from .parameterestimationcomposition import *
from .showgraph import *
from .report import *

__all__ = list(composition.__all__)
__all__.extend(pathway.__all__)
__all__.extend(compositionfunctionapproximator.__all__)
__all__.extend(parameterestimationcomposition.__all__)
__all__.extend(showgraph.__all__)
__all__.extend(report.__all__)



import doctest
import re

import graph_scheduler
import pytest

import psyneulink as pnl


@pytest.mark.parametrize("mod", [# Ports
                                 pnl.core.components.ports.parameterport,
                                 pnl.core.components.ports.outputport,
                                 pnl.core.components.ports.modulatorysignals.controlsignal,
                                 pnl.core.components.ports.modulatorysignals.gatingsignal,
                                 # Mechanisms
                                 # FIX 5/8/20 ELIMINATE SYSTEM [JDC] -- REFERENCES TO LABELS REQUIRE REFACTORING
                                 # pnl.core.components.mechanisms.mechanism,
                                 pnl.core.components.mechanisms.processing.transfermechanism,
                                 pnl.core.components.mechanisms.processing.integratormechanism,
                                 pnl.core.components.mechanisms.processing.objectivemechanism,
                                 pnl.core.components.mechanisms.modulatory.control.controlmechanism,
                                 # Functions
                                 pnl.core.components.functions.function,
                                 pnl.core.components.functions.stateful.memoryfunctions
                                ])
def test_core_docs(mod, capsys):
    fail, total = doctest.testmod(mod)
    if fail > 0:
        captured = capsys.readouterr()
        pytest.fail("{} out of {} examples failed:\n{}\n{}".format(
            fail, total, captured.err, captured.out), pytrace=False)

@pytest.mark.parametrize("mod", [# Mechanisms
                                 pnl.library.components.mechanisms.processing.integrator.ddm,
                                 pnl.library.components.mechanisms.processing.objective.comparatormechanism,
                                 pnl.library.components.mechanisms.processing.integrator.episodicmemorymechanism,
                                 # Scheduling
                                 pnl.core.scheduling.scheduler,
                                 # Logs
                                 pnl.core.globals.log,
                                ])
def test_other_docs(mod, capsys):
    fail, total = doctest.testmod(mod, optionflags=doctest.REPORT_NDIFF)
    if fail > 0:
        captured = capsys.readouterr()
        pytest.fail("{} out of {} examples failed:\n{}\n{}".format(
            fail, total, captured.err, captured.out), pytrace=False)


def consistent_doc_attrs(*objs):
    return (
        all(o.__doc__ is None for o in objs)
        or all(isinstance(o.__doc__, str) for o in objs)
    )


@pytest.mark.parametrize(
    'mod',
    [
        pnl.core.scheduling.scheduler,
        pnl.core.scheduling.condition,
        pnl.core.scheduling.time,
    ]
)
def test_scheduler_substitutions(mod):
    for cls, repls in mod._doc_subs.items():
        if cls is None:
            cls = mod
        else:
            cls = getattr(mod, cls)

        for pattern, repl in repls:
            # remove any group substitution strings
            assert re.sub(r'\\\d', '', repl) in cls.__doc__
            # where the pattern is being replaced, check that the pattern is gone
            if not re.match(r'\\\d', repl):
                assert not re.match(pattern, cls.__doc__)

    for pattern, repl in pnl.core.scheduling._global_doc_subs:
        for cls_name in mod.__all__:
            cls = getattr(mod, cls_name)
            try:
                ext_cls = getattr(graph_scheduler, cls_name)
            except AttributeError:
                continue

            assert consistent_doc_attrs(cls, ext_cls)
            # global replacements may not happen in every docstring
            if cls.__doc__ is not None:
                assert re.sub(r'\\\d', '', repl) in cls.__doc__ or not re.match(pattern, ext_cls.__doc__)

        ext_module = getattr(graph_scheduler, mod.__name__.split('.')[-1]).__doc__
        assert consistent_doc_attrs(mod, ext_module)
        assert re.sub(r'\\\d', '', repl) in mod.__doc__ or not re.match(pattern, ext_module.__doc__)



import numpy as np
import pytest

from psyneulink.core import llvm as pnlvm


@pytest.mark.llvm
@pytest.mark.parametrize('mode', ['CPU',
                                  pytest.param('PTX', marks=pytest.mark.cuda)])
@pytest.mark.parametrize('val', [np.uint8(0x7e),
                                 np.uint16(0x7eec),
                                 np.uint32(0x7eedbeee),
                                 np.uint64(0x7eedcafedeadbeee)
                                ], ids=lambda x: str(x.dtype))
def test_integer_broadcast(mode, val):
    custom_name = None
    with pnlvm.LLVMBuilderContext.get_current() as ctx:
        custom_name = ctx.get_unique_name("broadcast")
        int_ty = ctx.convert_python_struct_to_llvm_ir(val)
        int_array_ty = pnlvm.ir.ArrayType(int_ty, 8)
        func_ty = pnlvm.ir.FunctionType(pnlvm.ir.VoidType(),
                                        (int_ty.as_pointer(),
                                         int_array_ty.as_pointer()))
        function = pnlvm.ir.Function(ctx.module, func_ty, name=custom_name)

        i, o = function.args
        block = function.append_basic_block(name="entry")
        builder = pnlvm.ir.IRBuilder(block)
        ival = builder.load(i)
        ival = builder.add(ival, ival.type(1))
        with pnlvm.helpers.array_ptr_loop(builder, o, "broadcast") as (b, i):
            out_ptr = builder.gep(o, [ctx.int32_ty(0), i])
            builder.store(ival, out_ptr)
        builder.ret_void()

    binf = pnlvm.LLVMBinaryFunction.get(custom_name)
    val = np.asarray(val)
    res = binf.np_buffer_for_arg(1)

    if mode == 'CPU':
        binf(val, res)
    else:
        binf.cuda_wrap_call(val, res)

    assert all(res == np.broadcast_to(val + 1, 8))



"""
This implements a model of Locus Coeruleus / Norepinephrine (LC/NE) function described in `Gilzenrat et al. (2002)
<http://www.sciencedirect.com/science/article/pii/S0893608002000552?via%3Dihub>`_, used to simulate behavioral
and electrophysiological data (from LC recordings) in non-human primates.

"""
import argparse
import sys

import numpy as np
import psyneulink as pnl

parser = argparse.ArgumentParser()
parser.add_argument('--no-plot', action='store_false', help='Disable plotting', dest='enable_plot')
parser.add_argument(
    '--noise-stddev',
    type=float,
    help='Standard deviation of Gaussian noise distributions (default: 0.22, Gilzenrat et al. (2002))',
    default=0.22
)
args = parser.parse_args()

# Define Variables ----------------------------------------------------------------------------------------------------

# Weights & Biases:
b_decision = 0.00   # Bias on decision units (not biased)
b_response = 2.00   # Bias on response unit --- NOTE: Gilzenrat has negative signs in his logistic equation
w_XiIi = 1.00       # Connection weight from input units I1 and I2 to respective decision units X1 and X2
w_XiIj = 0.33       # Cross talk weight from input unit to opposite decision unit
w_XiXi = 1.00       # Recurrent self-connection weight for both decision units
w_XiXj = 1.00       # Magnitude of mutually inhibitory weight between decision units
w_X3X1 = 1.84       # Connection weight from target decision unit (X1) to response unit (X3)
w_X3X3 = 2.00       # Recurrent self-connection weight for the response unit
w_vX1 = 0.30        # Connection weight from target decision unit X1 to the abstracted LC

# Other parameters:
a = 0.50        # Parameter describing shape of the FitzHugh–Nagumo cubic nullcline for the fast excitation variable v
d = 0.50        # Baseline level of intrinsic, uncorrelated LC activity
G = 0.50        # Base level of gain applied to decision and response units
k = 3.00        # Scaling factor for transforming NE release (u) to gain (g) on potentiated units
SD = args.noise_stddev  # Standard deviation of Gaussian noise distributions | NOTE: 0.22 in Gilzenrat paper
tau_v = 0.05    # Time constant for fast LC excitation variable v | NOTE: tau_v is misstated in the Gilzenrat paper(0.5)
tau_u = 5.00    # Time constant for slow LC recovery variable (‘NE release’) u
dt = 0.02       # Time step size for numerical integration

# Switch between high C and low C conditions
high_C = True
if high_C:
    C = 0.95            # Mode ("coherence")
    initial_hv = 0.07   # Initial value for h(v)
    initial_u = 0.14      # initial value u
else:
    C = 0.55            # Mode ("coherence")
    initial_hv = 0.2    # Initial value for h(v)
    initial_u = 0.2     # initial value u

initial_v = (initial_hv - (1 - C) * d) / C    # get initial v from initial h(v)

# Create mechanisms ---------------------------------------------------------------------------------------------------

# Input Layer --- [ Target, Distractor ]
input_layer = pnl.TransferMechanism(
    input_shapes=2,
    initial_value=np.array([[0.0, 0.0]]),
    name='INPUT LAYER'
)

# Create Decision Layer  --- [ Target, Distractor ]

decision_layer = pnl.LCAMechanism(
    input_shapes=2,
    time_step_size=dt,
    leak=1.0,
    self_excitation=w_XiXi,
    competition=w_XiXj,
    #  Recurrent matrix: [  w_XiXi   -w_XiXj ]
    #                    [ -w_XiXj    w_XiXi ]
    function=pnl.Logistic(x_0=b_decision),
    noise=pnl.NormalDist(standard_deviation=SD),
    integrator_mode=True,
    name='DECISION LAYER'
)

# Create Response Layer  --- [ Target ]

response_layer = pnl.LCAMechanism(
    input_shapes=1,
    time_step_size=dt,
    leak=1.0,
    self_excitation=w_X3X3,
    #  Recurrent matrix: [w_X3X3]
    #  Competition param does not apply because there is only one unit
    function=pnl.Logistic(x_0=b_response),
    noise=pnl.NormalDist(standard_deviation=SD),
    integrator_mode=True,
    name='RESPONSE'
)

# Connect mechanisms --------------------------------------------------------------------------------------------------

# Weight matrix from Input Layer --> Decision Layer
input_weights = np.array([
    [w_XiIi, w_XiIj],
    [w_XiIj, w_XiIi]
])

# Weight matrix from Decision Layer --> Response Layer
output_weights = np.array([
    [w_X3X1],
    [0.00]
])

decision_pathway = pnl.Pathway(
    pathway=[
        input_layer,
        input_weights,
        decision_layer,
        output_weights,
        response_layer
    ],
    name='DECISION PROCESS'
)

# Monitor decision layer in order to modulate gain --------------------------------------------------------------------

LC = pnl.LCControlMechanism(
    integration_method="EULER",
    threshold_FitzHughNagumo=a,
    uncorrelated_activity_FitzHughNagumo=d,
    base_level_gain=G,
    scaling_factor_gain=k,
    time_step_size_FitzHughNagumo=dt,
    mode_FitzHughNagumo=C,
    time_constant_v_FitzHughNagumo=tau_v,
    time_constant_w_FitzHughNagumo=tau_u,
    a_v_FitzHughNagumo=-1.0,
    b_v_FitzHughNagumo=1.0,
    c_v_FitzHughNagumo=1.0,
    d_v_FitzHughNagumo=0.0,
    e_v_FitzHughNagumo=-1.0,
    f_v_FitzHughNagumo=1.0,
    a_w_FitzHughNagumo=1.0,
    b_w_FitzHughNagumo=-1.0,
    c_w_FitzHughNagumo=0.0,
    t_0_FitzHughNagumo=0.0,
    initial_v_FitzHughNagumo=initial_v,
    initial_w_FitzHughNagumo=initial_u,
    objective_mechanism=pnl.ObjectiveMechanism(
        function=pnl.Linear,
        monitor=[(
            decision_layer,
            None,
            None,
            np.array([
                [w_vX1],
                [0.0]]
            )
        )],
        name='LC ObjectiveMechanism'
    ),
    modulated_mechanisms=[decision_layer, response_layer],  # Modulate gain of decision & response layers
    name='LC'
)

task = pnl.Composition()
task.add_linear_processing_pathway(decision_pathway)
task.add_node(LC)

# This displays a diagram of the System
# task.show_graph()

# Create Stimulus -----------------------------------------------------------------------------------------------------

# number of trials
trials = 1000

# assign inputs to input_layer (Origin Mechanism) for each trial
stimulus_dictionary = {input_layer: np.repeat(np.array([[0.0, 0.0], [1.0, 0.0]]), 500, axis=0)}
# First 500 trials: target receives input of 0.0, distractor receives input of 0.0
# Second 500 trials: target receives input of 1.0, distractor receives input of 0.0

# Record results & run model ------------------------------------------------------------------------------------------

# Function to compute h(v) from LC's v value


def h_v(v, C, d):
    return C * v + (1 - C) * d


# Initialize output arrays for plotting and storing values
LC_results_h_of_v = [h_v(initial_v, C, d)]
LC_results_u = [initial_u]
decision_layer_target_values = [0.0]
decision_layer_distractor_values = [0.0]
response_layer_values = [0.0]


def record_trial(context):
    # After each trial, store all of the following values:
    LC_results_h_of_v.append(h_v(LC.parameters.value.get(context)[1][0], C, d))
    LC_results_u.append(LC.parameters.value.get(context)[2][0])
    decision_layer_target_values.append(decision_layer.parameters.value.get(context)[0][0])
    decision_layer_distractor_values.append(decision_layer.parameters.value.get(context)[0][1])
    response_layer_values.append(response_layer.parameters.value.get(context)[0][0])

    # Progress bar
    current_trial_num = len(LC_results_h_of_v)
    if current_trial_num % 50 == 0:
        percent = int(round((float(current_trial_num) / trials) * 100))
        sys.stdout.write("\r" + str(percent) + "% complete")
        sys.stdout.flush()


# Initialize progress bar
sys.stdout.write("\r0% complete")
sys.stdout.flush()

# Run the model
print('\nRunning model...')
task.run(
    inputs=stimulus_dictionary,
    num_trials=trials,
    call_after_trial=record_trial
)

# Plot results of all units into one figure ---------------------------------------------------------------------------
if args.enable_plot:
    import matplotlib.pyplot as plt

    print('\nModel run, generating plots...')

    # Create x axis "t" for plotting
    t = np.arange(0.0, 20.02, 0.02)

    # Plot target unit, distraction unit, response unit, h(v), and u using the values that were recorded after each trial
    plt.plot(
        t,
        decision_layer_target_values,
        label="target unit",
        color='green'
    )
    plt.plot(
        t,
        decision_layer_distractor_values,
        label="distraction unit",
        color='red'
    )
    plt.plot(
        t,
        response_layer_values,
        label="response unit",
        color='magenta'
    )
    plt.plot(
        t,
        pnl.convert_to_np_array(LC_results_h_of_v),
        label="h(v)",
        color='b'
    )
    plt.plot(
        t,
        pnl.convert_to_np_array(LC_results_u),
        label="u",
        color='black'
    )

    plt.xlabel('Time')
    plt.ylabel('Activation')
    plt.legend(loc='upper left')
    plt.xlim((0.0, 20.0))
    plt.ylim((-0.2, 1.2))
    x_values = [0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20]
    plt.xticks(x_values)
    plt.title('GILZENRAT 2002 PsyNeuLink', fontweight='bold')

    plt.show(block=not pnl._called_from_pytest)

    task.show_graph()
    print('\nPlots generated')



import versioneer

# Always prefer setuptools over distutils
from setuptools import find_packages, setup
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


def get_requirements(require_name=None):
    prefix = require_name + '_' if require_name is not None else ''
    with open(path.join(here, prefix + 'requirements.txt'), encoding='utf-8') as f:
        return f.read().strip().split('\n')


setup(
    name='psyneulink',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),

    description='A block modeling system for cognitive neuroscience',
    long_description=long_description,

    # Github address.
    url='https://github.com/PrincetonUniversity/PsyNeuLink',

    # Author details
    author='Jonathan Cohen, Princeton University',
    author_email='jdc@princeton.edu',

    # Choose your license
    license='Apache',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # This project is licensed as follows
        'License :: OSI Approved :: Apache Software License',

        # Supported Python Versions
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],

    # Require recent python
    python_requires=">=3.8",

    # What does your project relate to?
    keywords='cognitive modeling',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=get_requirements(),

    extras_require={
        'cuda': get_requirements('cuda'),
        'dev': get_requirements('dev'),
        'doc': get_requirements('doc'),
        'tutorial': get_requirements('tutorial'),
    }
)
